const meta = {
  docs: {
    description: 'enforce consistent variable types',
    category: 'Best Practices',
    recommended: true
  },
  fixable: null,
  schema: []
};

const create = function (context) {
  const TYPE_MAP = {};

  const getScopeId = (block) =>
    `${block.type}-${block.start}`;

  const findType = (scope, id, type) => {
    if (!scope || !scope.block) {
      return null;
    }

    if (TYPE_MAP[getScopeId(scope.block)] && id in TYPE_MAP[getScopeId(scope.block)]) {
      return TYPE_MAP[getScopeId(scope.block)][id];
    }

    return type !== 'MemberExpression' && findType(scope.upper, id);
  };


  const getId = (item) => {
    if (item.type === 'VariableDeclarator') {
      return item.id.name;
    }

    if (item.left) {
      if (item.left.type === 'MemberExpression') {
        let name = '';
        let member = item.left;

        while (member.property && member.object) {
          name = `${member.property.name}.${name}`;
          member = member.object;
        }

        return name;
      }

      return item.left.name;
    }
  };

  const getType = (expression) => {
    if (!expression) return null;
    switch (expression.type) {
      case 'Literal': {
        if (expression.value === null) {
          return null
        }
        return typeof expression.value;
      }
      case 'ObjectExpression': return 'object';
      case 'ArrayExpression': return 'array';
      case 'FunctionExpression':
      case 'ArrowFunctionExpression': return 'function';
      default: return null;
    }
  };

  return {
    AssignmentExpression(node) {
      const id = getId(node);
      const newType = getType(node.right);

      if (!id || !newType) return;

      const scope = context.getScope();
      const declaredType = findType(scope, id, node.left.type);

      if (!declaredType) {
        const { block } = scope;

        if (!TYPE_MAP[getScopeId(block)]) {
          TYPE_MAP[getScopeId(block)] = {};
        }

        TYPE_MAP[getScopeId(block)][id] = newType;

        return;
      }

      if (declaredType && declaredType !== newType) {
        context.report({
          node: node.right,
          message: `Can't assign type ${newType} to ${declaredType}.`,
        });
      }
    },

    VariableDeclaration(node) {
      const { block } = context.getScope();

      if (!TYPE_MAP[getScopeId(block)]) {
        TYPE_MAP[getScopeId(block)] = {};
      }

      node.declarations.forEach((declaration) => {
        const id = getId(declaration);
        TYPE_MAP[getScopeId(block)][id] = getType(declaration.init);
      });
    }
  };
};

module.exports = {
  meta,
  create,
};
